/*
Copyright (c) 2012, David Liu
All rights reserved. Read the license for more information.
*/

/* OS */
#if defined(_WIN32)
#define COS_WIN32
#elif defined(__linux)
#define COS_LINUX
#endif

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <fstream>
using namespace std;

/* Windows */
#if defined(COS_WIN32)
#define WIN32_LEAN_AND_MEAN 0
#include <windows.h>
#include <GL/gl.h>
#include <mmsystem.h>
#include <winsock2.h>
#include <ws2ipdef.h>
#pragma comment (lib,"opengl32.lib")
#pragma comment (lib,"ws2_32.lib")
#pragma comment (lib,"mswsock.lib")
#pragma comment (lib,"winmm.lib")
#pragma warning (disable:4996)
#pragma warning (disable:4244)
#pragma warning (disable:4800)

/* Linux */
#elif defined(COS_LINUX)
#include <sys/time.h>
#include <unistd.h>
#include <dirent.h>
#include <fnmatch.h>
#include <GL/glx.h>
#include <X11/X.h>
//#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#include <X11/xpm.h>
#endif

/* OpenGL */
#include <GL/glext.h>
#if defined(COS_WIN32)
#include <GL/wglext.h>
#endif

/* BASS */
#include "bass.h"
#if defined(COS_WIN32)
#pragma comment (lib,"bass.lib")
#endif

// Windows resources
#if defined(COS_WIN32)
#define C_ICON 1
#endif

// CEngine version
#define CENGINE_VERSION "0.0.1"

// Global definitions
#define CSIZE_MAXPATH 260
#define CSIZE_ERRSTR 50

// Tools
const double CPI=acos(-1.0);
#define CLoops(LOOP,MAX) for (LOOP=0;LOOP<MAX;LOOP++)
#define CLoopc(LOOP,START,MAX,ADD) for (LOOP=START;LOOP<MAX;LOOP+=ADD)
#define CLoopr(LOOP,START,MIN,SUBT) for (LOOP=START;LOOP>MIN;LOOP-=SUBT)

// CEngine data types
struct cmodel
{
	cmodel()
	{
		vertex=NULL;
		texcoord=NULL;
		normal=NULL;
		face=NULL;
		facecnt=vertexcnt=normalcnt=texcoordcnt=0;
	}
	~cmodel()
	{
		unsigned int i;
		// Clear memory
		if (vertex!=NULL)
		{
			CLoops(i,vertexcnt)
			{
				delete[] vertex[i];
			}
			delete[] vertex;
		}
		if (texcoord!=NULL)
		{
			CLoops(i,texcoordcnt)
			{
				delete[] texcoord[i];
			}
			delete[] texcoord;
		}
		if (normal!=NULL)
		{
			CLoops(i,normalcnt)
			{
				delete[] normal[i];
			}
			delete[] normal;
		}
		if (face!=NULL)
		{
			CLoops(i,facecnt)
			{
				if (face[i].vertexindices!=NULL)
					delete[] face[i].vertexindices;
				if (face[i].texindices!=NULL)
					delete[] face[i].texindices;
				if (face[i].normalindices!=NULL)
					delete[] face[i].normalindices;
			}
			delete[] face;
		}
	}
	struct cface
	{
		unsigned int*vertexindices;
		unsigned int*texindices;
		unsigned int*normalindices;
	};
	unsigned int vertexcnt;
	unsigned int texcoordcnt;
	unsigned int normalcnt;
	unsigned int facecnt;
	float**vertex; // xyzw
	float**texcoord; // uvw
	float**normal; // ijk
	cface*face;
};
typedef unsigned int ctexture;
struct cparticles
{
	cparticles()
	{
	}
	~cparticles()
	{
	}
	unsigned int count;
	unsigned int*texture;
	float**dir; // xyz
	float**dev; // xyz
};
struct cwindowattribs
{
	unsigned char type;
    char*wndname;
    char*wndtitle;
	bool customicon,customcursor;
};
#define CWIN_POPUP 0
#define CWIN_STANDARD 1

/* CEngine declaration */
class CEngine
{
    // Base variables
#if defined(COS_WIN32)
	HINSTANCE hInstance;
	HWND hWnd;
	HDC hDC;
	HGLRC hRC;
	MSG msg;
	static LRESULT CALLBACK CWndProc(HWND,UINT,WPARAM,LPARAM);
	LRESULT CProcWindow(UINT,WPARAM,LPARAM);
	unsigned long long freq;
#elif defined(COS_LINUX)
    Window fake,hWin;
    XVisualInfo*hVi;
    Colormap cMap;
    XSetWindowAttributes wAtt;
    GLXContext hCx;
    XEvent hEvent;
    int hScr;
    Pixmap hIcon;
    Cursor hide,hCursor;
#endif
	unsigned short maxwidth,maxheight;
	// GUI variables
	bool toggle,toggle1,toggle2;
	float rectx,recty;
	// OpenGL
	bool AF;
	float maxAF;
	// Functions called internally
	bool CSetupWindow(cwindowattribs*);
	bool CEnableOpenGL();
	bool CEnableSound();
	bool CSetupGUI();
	bool CEnableSockets();
public:
	// Window management/input
	bool key[258]/* including key[CLMOUSE]+key[CRMOUSE] */,focus;
	int mx,my;
	int width,height;
	int xpos,ypos;
#if defined(COS_WIN32)
	void CMoveWindow(int posx,int posy,int pwidth=-1,int pheight=-1)
	{
		if (pwidth>=0)
			width=pwidth;
        else
            width=maxwidth;
		if (pheight>=0)
			height=pheight;
        else
            height=maxheight;
		xpos=posx;
		ypos=posy;
		MoveWindow(hWnd,xpos,ypos,width,height,false);
	};
	void CToggleCursor(bool visible)
	{
		if (visible)
		{
			while (ShowCursor(true)<0)
			{ShowCursor(true);}
		}
		else
		{
			while (ShowCursor(false)>=0)
			{ShowCursor(false);}
		}
	}
	void CProcessMessage()
	{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	void CMsgBox(char*text,char*title) {MessageBox(NULL,text,title,0);}
	void CSetCursor(unsigned short x,unsigned short y) {SetCursorPos(x,y);}
	unsigned long long CGetTime()
	{
		LARGE_INTEGER time;
		QueryPerformanceCounter(&time);
		return time.QuadPart/freq;
	}
	void CSwapBuffers() {SwapBuffers(hDC);}
	int CSearchDir(const char*path,const char*format,char***filestr)
	{
		// Count files
		unsigned int i=1,loop;
		WIN32_FIND_DATA fdata;
		HANDLE hfind;
		strcpy(outerr,rootpath);
		strcat(outerr,path);
		hfind=FindFirstFile(strcat(outerr,format),&fdata);
		if (hfind==INVALID_HANDLE_VALUE)
		{
			if (GetLastError()==ERROR_FILE_NOT_FOUND)
				return 0;
			else
				return -1;
		}
		while (FindNextFile(hfind,&fdata))
			i++;
		// Allocate buffer
		(*filestr)=new char*[i];
		// Restart
		FindClose(hfind);
		hfind=FindFirstFile(outerr,&fdata);
		// Assign data and allocate
		loop=0;
		(*filestr)[0]=new char[strlen(fdata.cFileName)+1];
		strcpy((*filestr)[0],fdata.cFileName);
		while (FindNextFile(hfind,&fdata)!=0)
		{
			loop++;
			(*filestr)[loop]=new char[strlen(fdata.cFileName)+1];
			strcpy((*filestr)[loop],fdata.cFileName);
		}
		// Close
		FindClose(hfind);
		return i;
	}
	void CSleep(unsigned int us)
	{
		Sleep(us/1000);
	}
	void CFullScreen(bool fs)
	{
	}
	void CMinimize(bool mz)
	{
		if (mz)
			ShowWindow(hWnd,SW_MINIMIZE);
		else
			ShowWindow(hWnd,SW_RESTORE);
	}
#elif defined(COS_LINUX)
	Display*hDis;
	void CMoveWindow(int posx,int posy,int pwidth=-1,int pheight=-1)
	{
	    if (pwidth>=0)
			width=pwidth;
        else
            width=maxwidth;
		if (pheight>=0)
			height=pheight;
        else
            height=maxheight;
		xpos=posx;
		ypos=posy;
	    XMoveResizeWindow(hDis,hWin,xpos,ypos,width,height);
	}
    void CToggleCursor(bool visible)
	{
	    if (visible)
            XDefineCursor(hDis,hWin,hCursor);
	    else
            XDefineCursor(hDis,hWin,hide);
	}
	void CProcessMessage();
	void CMsgBox(char*text,char*title)
	{
	    Window msgbox;
	    msgbox=XCreateWindow(hDis,RootWindow(hDis,hVi->screen),maxwidth/2-50,maxheight/2-20,100,40,0,0,InputOutput,NULL,NULL,NULL);
	    XMapRaised(hDis,msgbox);
    }
	void CSetCursor(unsigned short x,unsigned short y) {XWarpPointer(hDis,None,hWin,0,0,width,height,x,y);}
	unsigned long long CGetTime()
	{
        timeval time;
        gettimeofday(&time,NULL);
        return (unsigned long long)time.tv_sec*1000000+time.tv_usec;
	}
	void CSwapBuffers() {glXSwapBuffers(hDis,hWin);}
	int CSearchDir(const char*path,const char*format,char***filestr)
	{
	    unsigned int i=0;
	    DIR*dir;
	    dirent*temp;
	    // Count files
	    dir=opendir(path);
	    if (dir==NULL)
            return -1;
        while ((temp=readdir(dir)))
        {
            if (fnmatch(format,temp->d_name,FNM_PERIOD)==0)
                i++;
        }
        closedir(dir);
        if (i==0)
            return 0;
        // Allocate
        (*filestr)=new char*[i];
	    // Store in filestr
	    dir=opendir(path);
	    if (dir==NULL)
            return -1;
        i=0;
        while ((temp=readdir(dir)))
        {
            if (fnmatch(format,temp->d_name,FNM_PERIOD)==0)
            {
                (*filestr)[i]=new char[strlen(temp->d_name)+1];
                strcpy((*filestr)[i],temp->d_name);
                i++;
            }
        }
        closedir(dir);
	    return i;
	}
	void CSleep(unsigned int us)
	{
		usleep(us);
	}
	void CFullScreen(bool fs)
	{
	}
	void CMinimize(bool mz)
	{
	}
#endif
	// Server framework
	char master[10]; // IPv6 and IPv4 support
	// Engine properties
	char rootpath[CSIZE_MAXPATH];
	// Error report
	char inerr[CSIZE_ERRSTR];
	char outerr[CSIZE_MAXPATH+60];
	// Engine
	bool CSetupEngine(cwindowattribs*attribs,int pwidth=-1,int pheight=-1,int posx=0,int posy=0) // Creates a fully functional CEngine
	{
		setlocale(LC_ALL,"C");
	    focus=true;
		toggle=toggle1=toggle2=false;
		memset(key,0,sizeof(key));
	    #if defined(COS_WIN32)
		hInstance=GetModuleHandle(NULL);
		maxwidth=GetSystemMetrics(SM_CXSCREEN);
		maxheight=GetSystemMetrics(SM_CYSCREEN);
		if (pwidth>=0)
			width=pwidth;
        else
            width=maxwidth;
		if (pheight>=0)
			height=pheight;
        else
            height=maxheight;
        xpos=posx;
        ypos=posy;
		GetCurrentDirectory(MAX_PATH,rootpath);
		LARGE_INTEGER freqs;
		if (!QueryPerformanceFrequency(&freqs))
		{
			CMsgBox("Unable to set the timer. Program will now close.","Error");
			return false;
		}
		freq=freqs.QuadPart/1000000;
		#elif defined(COS_LINUX)
        if (pwidth>=0)
			width=pwidth;
        else
            width=-1;
		if (pheight>=0)
			height=pheight;
        else
            height=-1;
        xpos=posx;
        ypos=posy;
        if (!getcwd(rootpath,CSIZE_MAXPATH))
        {
            CMsgBox("Unable to set the rootpath. Program will now close.","Error");
			return false;
        }
        #endif
        strcat(rootpath,"/");
		if (!CSetupWindow(attribs)||!CEnableOpenGL()||!CEnableSound()||!CEnableSockets()||!CSetupGUI())
			return false;
		return true;
	}
	void CCleanupEngine()
	{
	    BASS_Free();
	    #if defined(COS_WIN32)
		if (hRC)
		{
			if (!wglMakeCurrent(NULL,NULL))
				CMsgBox("Unable release the device context and rendering context.","Error");
			if (!wglDeleteContext(hRC))
				CMsgBox("Unable to release the rendering context.","Error");
			hRC=NULL;
		}
		if (hDC&&!ReleaseDC(hWnd,hDC))
		{
			CMsgBox("Unable to release the device context.","Error");
			hDC=NULL;
		}
		if (hWnd&&!DestroyWindow(hWnd))
		{
			CMsgBox("Unable to destroy the window.","Error");
			hWnd=NULL;
		}
		if (hInstance&&!UnregisterClass("Wnd",hInstance))
		{
			CMsgBox("Unable to clear the window data.","Error");
			hInstance=NULL;
		}
		#elif defined(COS_LINUX)
		if (hIcon!=NULL)
            XFreePixmap(hDis,hIcon);
        if (hide!=NULL)
            XFreeCursor(hDis,hide);
		if (hCursor!=NULL)
            XFreeCursor(hDis,hCursor);
        if (hCx)
        {
            if (!glXMakeCurrent(hDis,None,NULL))
                CMsgBox("Unable to release the rendering context.","Error");
            glXDestroyContext(hDis,hCx);
            hCx=NULL;
        }
        XCloseDisplay(hDis);
        #endif
	}
	// Sound
	bool CLoadSound(const char*,unsigned long*,bool=false);
	void CStartSound(unsigned long*);
	void CPauseSound(unsigned long*);
	void CStopSound(unsigned long*);
	void CSoundVol(const unsigned short);
	// OpenGL
	bool CCheckExtension(const char*);
	void CResize2D(unsigned int,unsigned int,double,double);
	void CResize3D(unsigned int,unsigned int,double);
	bool CLoadBMP(const char*path,ctexture&texnum,bool anisotropy=false,bool mipmap=true);
	bool CLoadTGA(const char*path,ctexture&texnum,bool anisotropy=false,bool mipmap=true,bool sprite=false);
	void CVerticalSync(bool on);
	// Socket
	// GUI
	void CPrintValue(const double value,unsigned char precision,float xpos,float ypos,const float color[3],const unsigned short fonttype);
	void CPrintText(const char*str,float xpos,float ypos,const float color[3],const unsigned short fonttype);
	void CScrollV(float xpos,float ypos,float length,float*bufptr,float minval,float range);
	void CScrollH(float xpos,float ypos,float length,float*bufptr,float minval,float range);
	void CInputbar(float xpos,float ypos,float length,char*str,const unsigned short strlength,bool*clicked,unsigned short*charpos,const float barcolor[3],const float textcolor[3]);
	bool CPushbutton(const char*str,float xpos,float ypos,const float buttoncolor[3],const float textcolor[3]);
	// Model
	bool CLoadOBJ(const char*path,cmodel&model);
	void CRenderModel(cmodel&model);
};

/* CEngine headers */
#include "CWindow.h"
#include "COpenGL.h"
#include "CSound.h"
#include "CSockets.h"
#include "CGUI.h"
#include "CVector.h"
/* */
