/*
Copyright (c) 2012, David Liu
All rights reserved. Read the license for more information.
*/

/* Fonts */
#define CFONT_GUI 0
#define CFONT_TEXT 1
#define CFONT_SUBT 2
#define CFONTS 3
#define CFONTTYPES 2
#define CFONTSET 95
ctexture CFonttex[CFONTTYPES];

/* GUI structures/data */
// Font
struct
{
	float width,height;
	unsigned char font;
}CText[CFONTS];

// Standard colours
const float CGuicol_s[3]={0.6f,0.6f,0.6f}, // Standard (grey)
CGuicol_b[3]={0.1f,0.1f,1.0f}, // Blue
CGuicol_g[3]={0.1f,1.0f,0.1f}, // Green
CGuicol_r[3]={1.0f,0.1f,0.1f}, // Red
CGuicol_y[3]={1.0f,1.0f,0.1f}, // Yellow
CGuicol_t[3]={0.6f,1.0f,1.0f}, // Turquoise
CGuicol_w[3]={1.0f,1.0f,1.0f}, // White
CGuicol_d[3]={0.0f,0.0f,0.0f}; // Dark/Black

// GUI masks
unsigned int CGuimask[1];

// OpenGL list start (CFONT*CFONTSET)
#define CLISTSTART 285

/* Functions */
bool CEngine::CSetupGUI()
{
	unsigned short loop1,loop2;
	float temp1,temp2;
	/* Text types */
	CText[CFONT_GUI].font=0;
	CText[CFONT_GUI].height=0.75f;
	CText[CFONT_GUI].width=0.25f;
	CText[CFONT_TEXT].font=0;
	CText[CFONT_TEXT].height=0.6f;
	CText[CFONT_TEXT].width=0.225f;
	CText[CFONT_SUBT].font=1;
	CText[CFONT_SUBT].height=1.25f;
	CText[CFONT_SUBT].width=0.5f;
	// Fonts
	char name[CFONTTYPES][21];
	strcpy(name[0],"engine/font_text.tga");
	strcpy(name[1],"engine/font_subt.tga");
	// Text
	CLoops(loop1,CFONTTYPES)
	{
		if (!CLoadTGA(name[loop1],CFonttex[loop1],false,false))
		{
			CMsgBox(strcat(outerr,": Unable to find/load this font texture."),"Error");
			return false;
		}
	}
	// Build the fonts
	glGenLists(CLISTSTART);
	CLoops(loop1,CFONTS)
	{
		glPushMatrix();
		CLoops(loop2,CFONTSET)
		{
			temp1=float(loop2%16)/16.0f;
			temp2=float(loop2/16)/8.0f;
			glNewList(loop2+CFONTSET*loop1,GL_COMPILE);
			glBegin(GL_QUADS);
			glTexCoord2f(temp1,1.0f-temp2-0.12f);glVertex2f(0.0f,0.0f);
			glTexCoord2f(temp1+0.0625f,1.0f-temp2-0.12f);glVertex2f(CText[loop1].width,0.0f);
			glTexCoord2f(temp1+0.0625f,1.0f-temp2-0.0075f);glVertex2f(CText[loop1].width,CText[loop1].height);
			glTexCoord2f(temp1,1.0f-temp2-0.0075f);glVertex2f(0.0f,CText[loop1].height);
			glEnd();
			glTranslatef(CText[loop1].width,0.0f,0.0f);
			glEndList();
		}
		glPopMatrix();
	}
	// GUI mask
	if (!CLoadTGA("engine/gui_mask.tga",CGuimask[0],false,false))
	{
		CMsgBox(strcat(outerr,": Unable to find/load this font texture."),"Error");
		return false;
	}
	return true;
}

inline void CEngine::CPrintText(const char*text,float xpos,float ypos,const float color[3]=CGuicol_w,const unsigned short texttype=CFONT_TEXT)
{
	glColor3fv(color);
	glBindTexture(GL_TEXTURE_2D,CFonttex[CText[texttype].font]);
	glPushMatrix();
	glTranslatef(xpos,ypos,0.0f);
	glListBase(CFONTSET*texttype-32);
	glCallLists(strlen(text),GL_UNSIGNED_BYTE,text);
	glPopMatrix();
}

inline void CEngine::CPrintValue(const double value,unsigned char decimals,float xpos,float ypos,const float color[3]=CGuicol_w,const unsigned short texttype=CFONT_TEXT)
{
	char de[2];
	sprintf(de,"%d",decimals);
	strcpy(inerr,"%");strcat(inerr,"1");strcat(inerr,".");strcat(inerr,de);strcat(inerr,"f");
	sprintf(outerr,inerr,value);
	glColor3fv(color);
	glBindTexture(GL_TEXTURE_2D,CFonttex[CText[texttype].font]);
	glPushMatrix();
	glTranslatef(xpos,ypos,0.0f);
	glListBase(CFONTSET*texttype-32);
	glCallLists(strlen(outerr),GL_UNSIGNED_BYTE,outerr);
	glPopMatrix();
}

bool CEngine::CPushbutton(const char*text,float x,float y,const float color[3]=CGuicol_b,const float tcolor[3]=CGuicol_w)
{
	// Variables
	bool touched=false,clicked=false;
	float pwidth=strlen(text)*CText[CFONT_GUI].width+0.2f;
	// Pusbutton
	float ratx=width/rectx,raty=height/recty;
	if (mx>=x*ratx&&
		mx<=(x+pwidth)*ratx&&
		my>=y*raty&&
		my<=(y+1.0f)*raty)
	{
		touched=true;
		if (key[CKEY_LMOUSE]&&!toggle)
		{
			clicked=toggle=true;
			CStartSound(&CChan[CSND_PUSH]);
		}
		else if (!key[CKEY_LMOUSE])
			toggle=false;
	}
	if (touched)
		glColor3f(0.7f,0.7f,0.7f);
	else
		glColor3fv(color);
	glBindTexture(GL_TEXTURE_2D,CGuimask[0]);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);glVertex2f(x,y);
	glTexCoord2i(1,0);glVertex2f(x+pwidth,y);
	glTexCoord2i(1,1);glVertex2f(x+pwidth,y+1.0f);
	glTexCoord2i(0,1);glVertex2f(x,y+1.0f);
	glEnd();
	CPrintText(text,x+0.1,y+0.1,tcolor,CFONT_GUI);
	return clicked;
}

void CEngine::CInputbar(float x,float y,float pwidth,char*textstr,const unsigned short sizestr,bool*clicked,unsigned short*pos,const float color[3]=CGuicol_w,const float tcolor[3]=CGuicol_d)
{
	// Variables
	unsigned short maxchar=floor((pwidth-0.2f)/CText[CFONT_GUI].width);
	if (sizestr<maxchar) // Possible buffer overflow
		maxchar=sizestr;
	bool aloud=false;
	// Inputbar
	float ratx=width/rectx,raty=height/recty;
	if (mx>=x*ratx&&
		mx<=(x+pwidth)*ratx&&
		my>=y*raty&&
		my<=(y+1.0f)*raty)
	{
		if (key[CKEY_LMOUSE])
			*clicked=true;
	}
	else
		*clicked=false;
	if (*clicked)
		glColor3f(0.9f,0.9f,0.9f);
	else
		glColor3fv(color);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);glVertex2f(x,y);
	glTexCoord2i(1,0);glVertex2f(x+pwidth,y);
	glTexCoord2i(1,1);glVertex2f(x+pwidth,y+1.0f);
	glTexCoord2i(0,1);glVertex2f(x,y+1.0f);
	glEnd();
	if (*pos*CText[CFONT_GUI].width+0.2f<=pwidth&&*pos<maxchar)
		aloud=true;
	if (*clicked)
	{
        if (aloud)
		{
			if (key[CKEY_0]||key[CKEY_1]||key[CKEY_2]||key[CKEY_3]||key[CKEY_4]||key[CKEY_5]||key[CKEY_6]||key[CKEY_7]||key[CKEY_8]||key[CKEY_9]||
				key[CKEY_A]||key[CKEY_B]||key[CKEY_C]||key[CKEY_D]||key[CKEY_E]||key[CKEY_F]||key[CKEY_G]||key[CKEY_H]||key[CKEY_I]||
				key[CKEY_J]||key[CKEY_K]||key[CKEY_L]||key[CKEY_M]||key[CKEY_N]||key[CKEY_O]||
				key[CKEY_P]||key[CKEY_Q]||key[CKEY_R]||key[CKEY_S]||key[CKEY_T]||key[CKEY_U]||key[CKEY_V]||key[CKEY_W]||key[CKEY_X]||key[CKEY_Y]||
				key[CKEY_Z]||
				key[CKEY_SEMICOLON]||key[CKEY_DIVIDE]||key[CKEY_TILDE]||key[CKEY_SPACE]||key[CKEY_RBRACKET]||key[CKEY_LBRACKET]||
				key[CKEY_SEPERATOR]||key[CKEY_PLUS]||key[CKEY_MINUS]||key[CKEY_COMMA]||key[CKEY_PERIOD]||key[CKEY_QUOTE])
			{
				if (!toggle)
				{
					int loop;
					CLoopr(loop,maxchar-1,*pos,1)
					{
						textstr[loop]=textstr[loop-1];
					}
					// Numbers
					if (key[CKEY_0]&&!key[CKEY_LSHIFT])
						textstr[*pos]='0';
					else if (key[CKEY_1]&&!key[CKEY_LSHIFT])
						textstr[*pos]='1';
					else if (key[CKEY_2]&&!key[CKEY_LSHIFT])
						textstr[*pos]='2';
					else if (key[CKEY_3]&&!key[CKEY_LSHIFT])
						textstr[*pos]='3';
					else if (key[CKEY_4]&&!key[CKEY_LSHIFT])
						textstr[*pos]='4';
					else if (key[CKEY_5]&&!key[CKEY_LSHIFT])
						textstr[*pos]='5';
					else if (key[CKEY_6]&&!key[CKEY_LSHIFT])
						textstr[*pos]='6';
					else if (key[CKEY_7]&&!key[CKEY_LSHIFT])
						textstr[*pos]='7';
					else if (key[CKEY_8])
						textstr[*pos]='8';
					else if (key[CKEY_9]&&!key[CKEY_LSHIFT])
						textstr[*pos]='9';
					// Letters
					else if (key[CKEY_A]&&!key[CKEY_LSHIFT])
						textstr[*pos]='a';
					else if (key[CKEY_B]&&!key[CKEY_LSHIFT])
						textstr[*pos]='b';
					else if (key[CKEY_C]&&!key[CKEY_LSHIFT])
						textstr[*pos]='c';
					else if (key[CKEY_D]&&!key[CKEY_LSHIFT])
						textstr[*pos]='d';
                    else if (key[CKEY_E]&&!key[CKEY_LSHIFT])
						textstr[*pos]='e';
					else if (key[CKEY_F]&&!key[CKEY_LSHIFT])
						textstr[*pos]='f';
					else if (key[CKEY_G]&&!key[CKEY_LSHIFT])
						textstr[*pos]='g';
					else if (key[CKEY_H]&&!key[CKEY_LSHIFT])
						textstr[*pos]='h';
					else if (key[CKEY_I]&&!key[CKEY_LSHIFT])
						textstr[*pos]='i';
					else if (key[CKEY_J]&&!key[CKEY_LSHIFT])
						textstr[*pos]='j';
					else if (key[CKEY_K]&&!key[CKEY_LSHIFT])
						textstr[*pos]='k';
					else if (key[CKEY_L]&&!key[CKEY_LSHIFT])
						textstr[*pos]='l';
					else if (key[CKEY_M]&&!key[CKEY_LSHIFT])
						textstr[*pos]='m';
					else if (key[CKEY_N]&&!key[CKEY_LSHIFT])
						textstr[*pos]='n';
					else if (key[CKEY_O]&&!key[CKEY_LSHIFT])
						textstr[*pos]='o';
					else if (key[CKEY_P]&&!key[CKEY_LSHIFT])
						textstr[*pos]='p';
					else if (key[CKEY_Q]&&!key[CKEY_LSHIFT])
						textstr[*pos]='q';
					else if (key[CKEY_R]&&!key[CKEY_LSHIFT])
						textstr[*pos]='r';
					else if (key[CKEY_S]&&!key[CKEY_LSHIFT])
						textstr[*pos]='s';
					else if (key[CKEY_T]&&!key[CKEY_LSHIFT])
						textstr[*pos]='t';
					else if (key[CKEY_U]&&!key[CKEY_LSHIFT])
						textstr[*pos]='u';
					else if (key[CKEY_V]&&!key[CKEY_LSHIFT])
						textstr[*pos]='v';
					else if (key[CKEY_W]&&!key[CKEY_LSHIFT])
						textstr[*pos]='w';
					else if (key[CKEY_X]&&!key[CKEY_LSHIFT])
						textstr[*pos]='x';
					else if (key[CKEY_Y]&&!key[CKEY_LSHIFT])
						textstr[*pos]='y';
					else if (key[CKEY_Z]&&!key[CKEY_LSHIFT])
						textstr[*pos]='z';
					else if (key[CKEY_A]&&key[CKEY_LSHIFT])
						textstr[*pos]='A';
					else if (key[CKEY_B]&&key[CKEY_LSHIFT])
						textstr[*pos]='B';
					else if (key[CKEY_C]&&key[CKEY_LSHIFT])
						textstr[*pos]='C';
					else if (key[CKEY_D]&&key[CKEY_LSHIFT])
						textstr[*pos]='D';
                    else if (key[CKEY_E]&&key[CKEY_LSHIFT])
						textstr[*pos]='E';
					else if (key[CKEY_F]&&key[CKEY_LSHIFT])
						textstr[*pos]='F';
					else if (key[CKEY_G]&&key[CKEY_LSHIFT])
						textstr[*pos]='G';
					else if (key[CKEY_H]&&key[CKEY_LSHIFT])
						textstr[*pos]='H';
					else if (key[CKEY_I]&&key[CKEY_LSHIFT])
						textstr[*pos]='I';
					else if (key[CKEY_J]&&key[CKEY_LSHIFT])
						textstr[*pos]='J';
					else if (key[CKEY_K]&&key[CKEY_LSHIFT])
						textstr[*pos]='K';
					else if (key[CKEY_L]&&key[CKEY_LSHIFT])
						textstr[*pos]='L';
					else if (key[CKEY_M]&&key[CKEY_LSHIFT])
						textstr[*pos]='M';
					else if (key[CKEY_N]&&key[CKEY_LSHIFT])
						textstr[*pos]='N';
					else if (key[CKEY_O]&&key[CKEY_LSHIFT])
						textstr[*pos]='O';
					else if (key[CKEY_P]&&key[CKEY_LSHIFT])
						textstr[*pos]='P';
					else if (key[CKEY_Q]&&key[CKEY_LSHIFT])
						textstr[*pos]='Q';
					else if (key[CKEY_R]&&key[CKEY_LSHIFT])
						textstr[*pos]='R';
					else if (key[CKEY_S]&&key[CKEY_LSHIFT])
						textstr[*pos]='S';
					else if (key[CKEY_T]&&key[CKEY_LSHIFT])
						textstr[*pos]='T';
					else if (key[CKEY_U]&&key[CKEY_LSHIFT])
						textstr[*pos]='U';
					else if (key[CKEY_V]&&key[CKEY_LSHIFT])
						textstr[*pos]='V';
					else if (key[CKEY_W]&&key[CKEY_LSHIFT])
						textstr[*pos]='W';
					else if (key[CKEY_X]&&key[CKEY_LSHIFT])
						textstr[*pos]='X';
					else if (key[CKEY_Y]&&key[CKEY_LSHIFT])
						textstr[*pos]='Y';
					else if (key[CKEY_Z]&&key[CKEY_LSHIFT])
						textstr[*pos]='Z';
					// Symbols
					else if (key[CKEY_SEMICOLON]&&!key[CKEY_LSHIFT])
						textstr[*pos]=';';
					else if (key[CKEY_SEMICOLON]&&key[CKEY_LSHIFT])
						textstr[*pos]=':';
					else if (key[CKEY_PERIOD])
						textstr[*pos]='.';
					else if (key[CKEY_COMMA])
						textstr[*pos]=',';
					else if (key[CKEY_SEPERATOR]&&!key[CKEY_LSHIFT])
						textstr[*pos]='/';
					else if (key[CKEY_SEPERATOR]&&key[CKEY_LSHIFT])
						textstr[*pos]='?';
					else if (key[CKEY_SPACE])
						textstr[*pos]=' ';
					else if (key[CKEY_1]&&key[CKEY_LSHIFT])
						textstr[*pos]='!';
					else if (key[CKEY_2]&&key[CKEY_LSHIFT])
						textstr[*pos]='@';
					else if (key[CKEY_3]&&key[CKEY_LSHIFT])
						textstr[*pos]='#';
					else if (key[CKEY_4]&&key[CKEY_LSHIFT])
						textstr[*pos]='$';
					else if (key[CKEY_5]&&key[CKEY_LSHIFT])
						textstr[*pos]='%';
					else if (key[CKEY_6]&&key[CKEY_LSHIFT])
						textstr[*pos]='^';
					else if (key[CKEY_7]&&key[CKEY_LSHIFT])
						textstr[*pos]='&';
					else if (key[CKEY_TILDE]&&!key[CKEY_LSHIFT])
						textstr[*pos]='`';
					else if (key[CKEY_TILDE]&&key[CKEY_LSHIFT])
						textstr[*pos]='~';
					// Brackets
					else if (key[CKEY_0]&&key[CKEY_LSHIFT])
						textstr[*pos]=')';
					else if (key[CKEY_9]&&key[CKEY_LSHIFT])
						textstr[*pos]='(';
					else if (key[CKEY_LBRACKET]&&!key[CKEY_LSHIFT])
						textstr[*pos]='[';
					else if (key[CKEY_LBRACKET]&&key[CKEY_LSHIFT])
						textstr[*pos]='{';
					else if (key[CKEY_RBRACKET]&&!key[CKEY_LSHIFT])
						textstr[*pos]=']';
					else if (key[CKEY_RBRACKET]&&key[CKEY_LSHIFT])
						textstr[*pos]='}';
					// Math
					else if (key[CKEY_PLUS]&&!key[CKEY_LSHIFT])
						textstr[*pos]='=';
					else if (key[CKEY_PLUS]&&key[CKEY_LSHIFT])
						textstr[*pos]='+';
					else if (key[CKEY_MINUS]&&!key[CKEY_LSHIFT])
						textstr[*pos]='-';
					else if (key[CKEY_MINUS]&&key[CKEY_LSHIFT])
						textstr[*pos]='_';
					(*pos)++;
					toggle=true;
				}
			}
			else
				toggle=false;
		}
		// Tools
		if (key[CKEY_BACKSPACE])
		{
			if (*pos>0&&!toggle1)
			{
				int loop;
				(*pos)--;
				CLoopc(loop,*pos,maxchar,1)
				{
					if (loop+1<maxchar)
						textstr[loop]=textstr[loop+1];
					else
						textstr[loop]='\0';
				}
				toggle1=true;
			}
		}
		else
			toggle1=false;
		if (key[CKEY_RIGHT]&&*pos<strlen(textstr)||key[CKEY_LEFT]&&*pos>0)
		{
			if (key[CKEY_RIGHT]&&!toggle2)
			{
				(*pos)++;
				toggle2=true;
			}
			else if (key[CKEY_LEFT]&&!toggle2)
			{
				(*pos)--;
				toggle2=true;
			}
		}
		else
			toggle2=false;
		glColor3f(0.0f,0.0f,0.0f);
		glBegin(GL_LINES);
		glVertex2f(x+0.1f+(*pos)*CText[CFONT_GUI].width,y+0.2f);
		glVertex2f(x+0.1f+(*pos)*CText[CFONT_GUI].width,y+0.8f);
		glEnd();
	}
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	CPrintText(textstr,x+0.1f,y+0.1f,tcolor,CFONT_GUI);
}

void CEngine::CScrollH(float x,float y,float pwidth,float*valptr,float minv,float vrange)
{
	if (pwidth<=0.0f)
		return;
	float pos,value,sav=x+pwidth;
	float ratx=width/rectx,raty=height/recty;
	if (mx>=(x-0.1f)*ratx&&
		mx<=(sav+0.1f)*ratx&&
		my>=y*raty&&
		my<=(y+0.5f)*raty&&key[CKEY_LMOUSE])
	{
		pos=mx*rectx/width;
		if (pos<x)
		{
			pos=x;
			*valptr=value=minv;
		}
		else if (pos>sav)
		{
			pos=sav;
			*valptr=value=minv+vrange;
		}
		else
			*valptr=value=minv+(pos-x)/pwidth*vrange;
	}
	else
	{
		value=*valptr;
		pos=(value-minv)/vrange*pwidth+x;
	}
	glBindTexture(GL_TEXTURE_2D,CGuimask[0]);
	// Line
	glColor3f(0.45f,0.65f,0.4f);
	glBegin(GL_QUADS);
	glTexCoord2f(0.4f,0.0f);glVertex2f(x,y+0.3f);glTexCoord2f(0.4f,1.0f);glVertex2f(x,y+0.2f);
	glTexCoord2f(0.6f,1.0f);glVertex2f(sav,y+0.2f);glTexCoord2f(0.6f,0.0f);glVertex2f(sav,y+0.3f);
	glEnd();
	// Box
	glColor3f(0.7f,0.7f,0.7f);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);glVertex2f(pos-0.1f,y);glTexCoord2i(0,1);glVertex2f(pos+0.1f,y);
	glTexCoord2i(1,1);glVertex2f(pos+0.1f,y+0.5f);glTexCoord2i(1,0);glVertex2f(pos-0.1f,y+0.5f);
	glEnd();
}

void CEngine::CScrollV(float x,float y,float pheight,float*valptr,float minv,float vrange)
{
	if (pheight<=0.0f)
		return;
	float pos,value,sav=y+pheight;
	float ratx=width/rectx,raty=height/recty;
	if (mx>=x*ratx&&
		mx<=(x+0.5f)*ratx&&
		my>=(y-0.1f)*raty&&
		my<=(sav+0.1f)*raty&&key[CKEY_LMOUSE])
	{
		pos=my*recty/height;
		if (pos<y)
		{
			pos=y;
			*valptr=value=minv;
		}
		else if (pos>sav)
		{
			pos=sav;
			*valptr=value=minv+vrange;
		}
		else
			*valptr=value=minv+(pos-y)/pheight*vrange;
	}
	else
	{
		value=*valptr;
		pos=(value-minv)/vrange*pheight+y;
	}
	glBindTexture(GL_TEXTURE_2D,CGuimask[0]);
	// Line
	glColor3f(0.45f,0.65f,0.4f);
	glBegin(GL_QUADS);
	glTexCoord2f(0.4f,0.0f);glVertex2f(x+0.2f,y);glTexCoord2f(0.4f,1.0f);glVertex2f(x+0.3f,y);
	glTexCoord2f(0.6f,1.0f);glVertex2f(x+0.3f,sav);glTexCoord2f(0.6f,0.0f);glVertex2f(x+0.2f,sav);
	glEnd();
	// Box
	glColor3f(0.7f,0.7f,0.7f);
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);glVertex2f(x,pos-0.1f);glTexCoord2i(1,0);glVertex2f(x+0.5f,pos-0.1f);
	glTexCoord2i(1,1);glVertex2f(x+0.5f,pos+0.1f);glTexCoord2i(0,1);glVertex2f(x,pos+0.1f);
	glEnd();
}
