/*
Copyright (c) 2012, David Liu
All rights reserved. Read the license for more information.
*/

/* Macros */
#define CFrame(CODE) glClear(GL_COLOR_BUFFER_BIT);glLoadIdentity();glTranslatef(0.0f,0.0f,-1.0f);CODE

/* Additional functions */

/* Extension loader */
#define EXTENSIONS 2
enum{AF,MS};
bool CGlext[EXTENSIONS];
float maxAF;

/* Functions */
bool CEngine::CCheckExtension(const char*name)
{
	if (strstr((char*)glGetString(GL_EXTENSIONS),name)!=NULL)
		return true;
	return false;
}

void CEngine::CResize2D(unsigned int w,unsigned int h,double xrect=20.0,double yrect=20.0)
{
	width=w;height=h;
	rectx=xrect;recty=yrect;
	glViewport(0,0,width,height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0,xrect,0.0,yrect,0.0,1.0);
	glMatrixMode(GL_MODELVIEW);
}

void CEngine::CResize3D(unsigned int w,unsigned int h,double fov)
{
	width=w;height=h;
	glViewport(0,0,width,height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-tan(fov/360.0*CPI)*0.01,tan(fov/360.0*CPI)*0.01,-tan(fov/360.0*CPI)*0.01,tan(fov/360.0*CPI)*0.01,0.01,200.0);
	glMatrixMode(GL_MODELVIEW);
}

#if defined(COS_WIN32)
bool CEngine::CEnableOpenGL()
{
	PIXELFORMATDESCRIPTOR pfd;
	int iFormat;
	if (!(hDC=GetDC(hWnd)))
	{
		CMsgBox("Unable to create a device context. Program will now close.","Error");
		return false;
	}
	ZeroMemory(&pfd,sizeof(pfd));
	pfd.nSize=sizeof(pfd);
	pfd.nVersion=1;
	pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType=PFD_TYPE_RGBA;
	pfd.cColorBits=24;
	pfd.cDepthBits=16;
	pfd.iLayerType=PFD_MAIN_PLANE;
	if (!(iFormat=ChoosePixelFormat(hDC,&pfd)))
	{
		CMsgBox("Unable to find a suitable pixel format. Program will now close.","Error");
		return false;
	}
	if (!SetPixelFormat(hDC,iFormat,&pfd))
	{
		CMsgBox("Unable to initialize the pixel formats. Program will now close.","Error");
		return false;
	}
	if (!(hRC=wglCreateContext(hDC)))
	{
		CMsgBox("Unable to create a rendering context. Program will now close.","Error");
		return false;
	}
	if (!wglMakeCurrent(hDC,hRC))
	{
		CMsgBox("Unable to activate the rendering context. Program will now close.","Error");
		return false;
	}
	// Extensions
	if ((AF=CCheckExtension("GL_EXT_texture_filter_anisotropic")))
		glGetFloatv(GL_TEXTURE_MAX_ANISOTROPY_EXT,&maxAF);
	 // CEngine OpenGL states
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glLineWidth(1.0f);
	glClearColor(0,0,0,1);
	return true;
}

void CEngine::CVerticalSync(bool enable)
{
	if (CCheckExtension("WGL_EXT_swap_control"))
	{
		PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT;
		wglSwapIntervalEXT=(PFNWGLSWAPINTERVALEXTPROC)wglGetProcAddress("wglSwapIntervalEXT");
		wglSwapIntervalEXT(enable);
	}
}
#elif defined(COS_LINUX)
bool CEngine::CEnableOpenGL()
{
    // Extensions
	if ((AF=CCheckExtension("GL_EXT_texture_filter_anisotropic")))
		glGetFloatv(GL_TEXTURE_MAX_ANISOTROPY_EXT,&maxAF);
     // Engine OpenGL states
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glClearColor(0,0,0,1);
    return true;
}

void CEngine::CVerticalSync(bool enable)
{
	if (CCheckExtension("GLX_EXT_swap_control"));
		//glXSwapIntervalEXT(hDis,GLX_SWAP_INTERVAL_EXT,(int)enable);
}
#endif

bool CEngine::CLoadBMP(const char*path,ctexture&texnum,bool anisotropy,bool mipmap)
{
    unsigned short int ftype;
    long int fbits,imgsize,i;
    short int planes,bitcnt;
    unsigned char sav;
    int width,height;
    unsigned char*data;
    FILE*f;
	strcpy(outerr,rootpath);
	f=fopen(strcat(outerr,path),"rb");
	// File reading
	if (f==NULL)
	{
	    strcpy(inerr,"Unable to find/open the bitmap.");
	    return false;
	}
    if(!fread(&ftype,sizeof(short int),1,f))
    {
        strcpy(inerr,"Unable to read the bitmap.");
        return false;
    }
    if (ftype!=19778)
    {
        strcpy(inerr,"Image file is not a bitmap file.");
        return false;
    }
    fseek(f,8,SEEK_CUR); // Skip header and get to the actual image data
    if (!fread(&fbits,sizeof(long int),1,f))
    {
        strcpy(inerr,"Unable to successfully read the bitmap data.");
        return false;
    }
    fseek(f,4,SEEK_CUR); // Skip part of header
    fread(&width,sizeof(int),1,f);
    fread(&height,sizeof(int),1,f);
    fread(&planes,sizeof(short int),1,f);
    if (planes!=1)
    {
        strcpy(inerr,"Number of planes do not equal 1.");
        return false;
    }
    if (!fread(&bitcnt,sizeof(short int),1,f)) // Bits per pixel
    {
        strcpy(inerr,"Unable to successfully read the bitmap data.");
        return false;
    }
    if (bitcnt!=24)
    {
        strcpy(inerr,"Bits per pixel do not equal 24.");
        return false;
    }
    imgsize=width*height*3;
    data=new unsigned char[imgsize];
    // Actual image data
    fseek(f,fbits,SEEK_SET);
    if (!fread(data,imgsize,1,f))
    {
		delete[] data;
        strcpy(inerr,"Unable to successfully read the bitmap data.");
        return false;
    }
    fclose(f);
    // Swap R and B
    CLoopc(i,0,imgsize,3)
    {
        sav=data[i];
        data[i]=data[i+2];
        data[i+2]=sav;
    }
    // Generate texture
    glGenTextures(1,&texnum);
    glBindTexture(GL_TEXTURE_2D,texnum);
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, mipmap ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP,GL_TRUE);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    if (AF&&anisotropy)
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,maxAF);
    glTexImage2D(GL_TEXTURE_2D,0,3,width,height,0,GL_RGB,GL_UNSIGNED_BYTE,data);
	delete[] data;
    return true;
}

bool CEngine::CLoadTGA(const char*path,ctexture&texnum,bool anisotropy,bool mipmap,bool sprite)
{
	unsigned int loop1,loop2,loop3;
	FILE*f;
	unsigned char tgacmp[12];
	unsigned char tgainfo[6];
	unsigned char tgaheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};
	unsigned char*tgasize;
	unsigned int tgadim[2]; // Width and height
	strcpy(outerr,rootpath);
	f=fopen(strcat(outerr,path),"rb");
	if (f==NULL)
	{
		strcpy(inerr,"Unable to find/open the tga.");
		return false;
	}
	if (fread(tgacmp,1,sizeof(tgacmp),f)==sizeof(tgacmp)&&memcmp(tgaheader,tgacmp,sizeof(tgaheader))==0&&fread(tgainfo,1,sizeof(tgainfo),f)==sizeof(tgainfo))
	{
		tgadim[0]=tgainfo[1]*256+tgainfo[0];
		tgadim[1]=tgainfo[3]*256+tgainfo[2];
		if (tgadim[0]<=0||tgadim[1]<=0||tgainfo[4]!=32)
		{
			fclose(f);
			return false;
		}
		loop1=tgadim[0]*tgadim[1]*4; // Size
		tgasize=new unsigned char[loop1];
		if (fread(tgasize,1,loop1,f)!=loop1)
		{
			delete[] tgasize;
			fclose(f);
			strcpy(inerr,"Failed to successfully read the tga data.");
			return false;
		}
		fclose(f);
		CLoopc(loop2,0,loop1,4)
		{
			loop3=tgasize[loop2];
			tgasize[loop2]=tgasize[loop2+2];
			tgasize[loop2+2]=loop3;
		}
		// Build texture
		glGenTextures(1,&texnum);
		glBindTexture(GL_TEXTURE_2D,texnum);
		if (sprite)
			glTexEnvi(GL_POINT_SPRITE,GL_COORD_REPLACE,GL_TRUE);
		glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, mipmap ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR );
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP,GL_TRUE);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		if (AF&&anisotropy)
			glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,maxAF);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,tgadim[0],tgadim[1],0,GL_RGBA,GL_UNSIGNED_BYTE,tgasize);
		delete[] tgasize;
		return true;
	}
	strcpy(inerr,"Incompatible tga header.");
	return false;
}

bool CEngine::CLoadOBJ(const char*path,cmodel&mdl)
{
	// Init
	unsigned int i,v,t,n,c=0,b;
	char buf[256];
	strcpy(outerr,rootpath);
	FILE*f=fopen(strcat(outerr,path),"rb");
	if (f==NULL)
	{
		strcpy(inerr,"Failed to find/open the obj file.");
		return false;
	}
	// Count objects
	while (fgets(buf,256,f)!=NULL&&feof(f)==0)
    {
		if (buf[0]!='#')
		{
			if (buf[0]=='v')
			{
				if (buf[1]==' ')
					mdl.vertexcnt++;
			   else if (buf[1]=='t')
					mdl.texcoordcnt++;
				else if (buf[1]=='n')
					mdl.normalcnt++;
				else
				{
					strcpy(inerr,"Error found in obj file.");
					return false;
				}
			}
			else if (buf[0]=='f')
			{
				if (sscanf(buf+2,"%d/%d/%d",&v,&n,&t)==3)
				{
					if (mdl.texcoordcnt==0&&mdl.normalcnt==0)
					{
						strcpy(inerr,"Missing texture coordinates and normals.");
						return false;
					}
					mdl.facecnt++;
				}
				else if (sscanf(buf+2,"%d//%d",&v,&n)==2)
				{
					if (mdl.normalcnt==0)
					{
						strcpy(inerr,"Missing normals.");
						return false;
					}
					mdl.facecnt++;
				}
				else if (sscanf(buf+2,"%d/%d",&v,&t)==2)
				{
					if (mdl.texcoordcnt==0)
					{
						strcpy(inerr,"Missing texture coordinates.");
						return false;
					}
					mdl.facecnt++;
				}
				else if (sscanf(buf+2,"%d",&v)==1)
					mdl.facecnt++;
				else
				{
					strcpy(inerr,"Error found in obj file.");
					return false;
				}
			}
		}
	}
	if (mdl.vertexcnt==0)
	{
		strcpy(inerr,"No vertices found in obj file.");
		return false;
    }
	// Allocate data
	mdl.vertex=new float*[mdl.vertexcnt];
	CLoops(i,mdl.vertexcnt)
	{
		mdl.vertex[i]=new float[4];
	}
	if (mdl.texcoordcnt>0)
    {
		mdl.texcoord=new float*[mdl.texcoordcnt];
		CLoops(i,mdl.texcoordcnt)
		{
			mdl.texcoord[i]=new float[3];
		}
    }
	if (mdl.normalcnt>0)
    {
		mdl.normal=new float*[mdl.normalcnt];
		CLoops(i,mdl.normalcnt)
		{
			mdl.normal[i]=new float[3];
		}
    }
	if (mdl.facecnt>0)
		mdl.face=new cmodel::cface[mdl.facecnt];
	CLoops(i,mdl.facecnt)
	{
		mdl.face[i].vertexindices=NULL;
		mdl.face[i].texindices=NULL;
		mdl.face[i].normalindices=NULL;
	}
	// Store data
	v=t=n=0;
	rewind(f);
	while (fgets(buf,256,f)!=NULL&&!feof(f))
	{
		if (buf[0]!='#')
		{
			if (buf[0]=='v')
			{
				if (buf[1]==' ')
				{
					if (sscanf(buf+2,"%f %f %f %f",&mdl.vertex[v][0],&mdl.vertex[v][1],&mdl.vertex[v][2],&mdl.vertex[v][3])!=4)
					{
						if (sscanf(buf+2,"%f %f %f",&mdl.vertex[v][0],&mdl.vertex[v][1],&mdl.vertex[v][2])!=3)
						{
							strcpy(inerr,"Error found in obj file.");
							return false;
						}
						else
							mdl.vertex[v][3]=1.0f;
					}
					v++;
				}
				else if (buf[1]=='t')
				{
					if (sscanf(buf+2, "%f %f %f",&mdl.texcoord[t][0],&mdl.texcoord[t][1],&mdl.texcoord[t][2])!=3)
					{
						if (sscanf(buf+2,"%f %f",&mdl.texcoord[t][0],&mdl.texcoord[t][1])!=2)
						{
							if (sscanf(buf+2,"%f",&mdl.texcoord[t][0])!=1)
							{
								strcpy(inerr,"Error found in obj file.");
								return false;
							}
							else
							{
								mdl.texcoord[t][1]=0.0f;
								mdl.texcoord[t][2]=0.0f;
							}
						}
						else
							mdl.texcoord[t][2]=0.0f;
					}
					t++;
				}
				else if (buf[1]=='n')
				{
					if (sscanf(buf+2,"%f %f %f",&mdl.normal[n][0],&mdl.normal[n][1],&mdl.normal[n][2])!=3)
					{
						strcpy(inerr,"Error found in obj file.");
						return false;
					}
					n++;
				}
			}
			else if (buf[0]=='f')
			{
				mdl.face[c].vertexindices=new unsigned int[3];
				if (mdl.texcoordcnt>0)
					mdl.face[c].texindices=new unsigned int[3];
				if (mdl.normalcnt>0)
					mdl.face[c].normalindices=new unsigned int[3];
				b=1; // Counter
				CLoops(i,3)
				{
					b++;
					if (sscanf(buf+b,"%d/%d/%d",&mdl.face[c].vertexindices[i],&mdl.face[c].texindices[i],&mdl.face[c].normalindices[i])!=3)
					{
						if (sscanf(buf+b,"%d//%d",&mdl.face[c].vertexindices[i],&mdl.face[c].normalindices[i])!=2)
						{
							if (sscanf(buf+b,"%d/%d",&mdl.face[c].vertexindices[i],&mdl.face[c].texindices[i])!=2)
								sscanf(buf+b,"%d",&mdl.face[c].vertexindices[i]);
						}
					}
					mdl.face[c].vertexindices[i]--;
					if (mdl.texcoordcnt>0)
						mdl.face[c].texindices[i]--;
					if (mdl.normalcnt>0)
						mdl.face[c].normalindices[i]--;
					while (b<strlen(buf)&&buf[b]!=' ')
						b++;
				}
				c++;
			}
		}
	}
	return true;
}

inline void CEngine::CRenderModel(cmodel&mdl)
{
	unsigned int i,j;
	glBegin(GL_TRIANGLES);
	CLoops(i,mdl.facecnt)
	{
		CLoops(j,3)
		{
			if (mdl.texcoordcnt>0)
				glTexCoord3fv(mdl.texcoord[mdl.face[i].texindices[j]]);
			if (mdl.normalcnt>0)
				glNormal3fv(mdl.normal[mdl.face[i].normalindices[j]]);
			glVertex4fv(mdl.vertex[mdl.face[i].vertexindices[j]]);
		}
	}
	glEnd();
}