/*
Copyright (c) 2012, David Liu
All rights reserved. Read the license for more information.
*/

/* Primitives */
struct CVector
{
	CVector() {};
	CVector(float tx,float ty,float tz) : x(tx),y(ty),z(tz) {};
	CVector&operator+=(CVector other)
	{
		CVector&tmp(*this);
		tmp.x+=other.x;
		tmp.y+=other.y;
		tmp.z+=other.z;
		return tmp;
	}
	CVector&operator-=(CVector other)
	{
		CVector&tmp(*this);
		tmp.x-=other.x;
		tmp.y-=other.y;
		tmp.z-=other.z;
		return tmp;
	}
	CVector&operator/=(float denom)
	{
		CVector&tmp(*this);
		tmp.x/=denom;
		tmp.y/=denom;
		tmp.z/=denom;
		return tmp;
	}
	CVector&operator*=(float factor)
	{
		CVector&tmp(*this);
		tmp.x*=factor;
		tmp.y*=factor;
		tmp.z*=factor;
		return tmp;
	}
	CVector operator*(float factor)
	{
		CVector tmp;
		tmp.x=this->x*factor;
		tmp.y=this->y*factor;
		tmp.z=this->z*factor;
		return tmp;
	}
	bool operator!=(CVector other)
	{
		CVector&tmp(*this);
		if (tmp.x!=other.x||tmp.y!=other.y||tmp.z!=other.z)
			return true;
		else
			return false;
	}
	float magnitude()
	{
		CVector&tmp(*this);
		return sqrt(tmp.x*tmp.x+tmp.y*tmp.y+tmp.z*tmp.z);
	}
	void normalize()
	{
		CVector&tmp(*this);
		float divider=tmp.magnitude();
		if (divider!=1.0f)
			tmp/=divider;
	}
	float x,y,z;
};
const CVector CNullvect(0.0f,0.0f,0.0f);

struct CVertex
{
	CVertex() {};
	CVertex(float tx,float ty,float tz) : x(tx),y(ty),z(tz) {};
	CVertex&operator+=(CVector other)
	{
		CVertex&tmp(*this);
		tmp.x+=other.x;
		tmp.y+=other.y;
		tmp.z+=other.z;
		return tmp;
	}
	CVertex&operator-=(CVector other)
	{
		CVertex&tmp(*this);
		tmp.x-=other.x;
		tmp.y-=other.y;
		tmp.z-=other.z;
		return tmp;
	}
	CVertex operator+(CVector other)
	{
		CVertex tmp;
		tmp.x=other.x+this->x;
		tmp.y=other.y+this->y;
		tmp.z=other.z+this->z;
		return tmp;
	}
	CVertex operator-(CVector other)
	{
		CVertex tmp;
		tmp.x=other.x-this->x;
		tmp.y=other.y-this->y;
		tmp.z=other.z-this->z;
		return tmp;
	}
	CVector operator+(CVertex other)
	{
		CVector tmp;
		tmp.x=other.x+this->x;
		tmp.y=other.y+this->y;
		tmp.z=other.z+this->z;
		return tmp;
	}
	CVector operator-(CVertex other)
	{
		CVector tmp;
		tmp.x=other.x-this->x;
		tmp.y=other.y-this->y;
		tmp.z=other.z-this->z;
		return tmp;
	}
	float x,y,z;
};

/* Basic objects */
struct CLine
{
	CVertex pos;
	CVector dir;
	float length;
};

struct CRect
{
	CVertex pos;
	CVector normal;
	CVector dir1,dir2;
};

struct CSphere
{
	CVertex pos;
	float radius;
};

/* Operations */
inline float CDotProduct(CVector one,CVector two) {return one.x*two.x+one.y*two.y+one.z*two.z;}

inline CVector CCrossProduct(CVector one,CVector two)
{
	CVector tmp(one.y*two.z-one.z*two.y,one.z*two.x-one.x*two.z,one.x*two.y-one.y*two.x);
	return tmp;
}

inline float CVectorAngle(CVector one,CVector two) {return acos(CDotProduct(one,two)/one.magnitude()/two.magnitude());}

inline float CDistance(CVertex vector,CVertex other)
{
    float tempx=other.x-vector.x,tempy=other.y-vector.y,tempz=other.z-vector.z;
    return sqrt(tempx*tempx+tempy*tempy+tempz*tempz);
}

float cvtmp;

/* Collision detection */
bool CVertex_Rect(CVertex v,CRect R)
{
	float labda,nu;
	// Express labda in nu
	// R.pos.x+R.dir1.x*labda+R.dir2.x*nu=v.x;
	// R.pos.y+R.dir1.y*labda+R.dir2.y*nu=v.y;
	// labda=(v.x-R.pos.x-R.dir2.x*nu)/R.dir1.x
	// Get nu
	// nu=(v.y-R.pos.y-R.dir1.y*labda)/R.dir2.y
	// nu=(v.y-R.pos.y-R.dir1.y*((v.x-R.pos.x-R.dir2.x*nu)/R.dir1.x))/R.dir2.y
	// nu=v.y/R.dir2.y-R.pos.y/R.dir2.y-v.x/R.dir1.x/R.dir2.y*R.dir1.y-R.pos.x/R.dir1.x/R.dir2.y*R.dir1.y-R.dir2.x/R.dir1.x/R.dir2.y*R.dir1.y*nu
	// R.dir2.x/R.dir1.x/R.dir2.y*R.dir1.y is the coefficient
	// (1.0f+R.dir2.x/R.dir1.x/R.dir2.y*R.dir1.y)*nu=v.y/R.dir2.y-R.pos.y/R.dir2.y-v.x/R.dir1.x/R.dir2.y*R.dir1.y-R.pos.x/R.dir1.x/R.dir2.y*R.dir1.y
	//if (R.dir2.y!=0.0f&&R.dir1.x!=0.0f) // Prevent div by 0
	//else if (R.dir2.z!=0.0f&&R.dir1.z!=0.0f)
	//	nu=(v.y/R.dir2.y-R.pos.y/R.dir2.y-v.x/R.dir1.x/R.dir2.y*R.dir1.y-R.pos.x/R.dir1.x/R.dir2.y*R.dir1.y)/(1.0f+R.dir2.x/R.dir1.x/R.dir2.y*R.dir1.y)
	//else
		//nu=(v.y/R.dir2.y-R.pos.y/R.dir2.y-v.x/R.dir1.x/R.dir2.y*R.dir1.y-R.pos.x/R.dir1.x/R.dir2.y*R.dir1.y)/(1.0f+R.dir2.x/R.dir1.x/R.dir2.y*R.dir1.y)
	// Get labda
	// labda=(v.x-R.pos.x-R.dir2.x*nu)/R.dir1.x;
	// Final optimized version:
	nu=(v.z/R.dir2.x-R.pos.z/R.dir2.x-v.x/R.dir1.x/R.dir2.x*R.dir1.z-R.pos.x/R.dir1.x/R.dir2.x*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.x/R.dir2.x*R.dir1.z);
	if (nu!=nu)
	{
		nu=(v.y/R.dir2.y-R.pos.z/R.dir2.y-v.x/R.dir1.x/R.dir2.y*R.dir1.z-R.pos.x/R.dir1.x/R.dir2.y*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.x/R.dir2.y*R.dir1.z);
		if (nu!=nu)
	{
		nu=(v.z/R.dir2.z-R.pos.z/R.dir2.z-v.x/R.dir1.x/R.dir2.z*R.dir1.z-R.pos.x/R.dir1.x/R.dir2.z*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.x/R.dir2.z*R.dir1.z);
		if (nu!=nu)
	{
		nu=(v.z/R.dir2.x-R.pos.z/R.dir2.x-v.x/R.dir1.y/R.dir2.x*R.dir1.z-R.pos.x/R.dir1.y/R.dir2.x*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.y/R.dir2.x*R.dir1.z);
		if (nu!=nu)
	{
		nu=(v.y/R.dir2.y-R.pos.z/R.dir2.y-v.x/R.dir1.y/R.dir2.y*R.dir1.z-R.pos.x/R.dir1.y/R.dir2.y*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.y/R.dir2.y*R.dir1.z);
		if (nu!=nu)
	{
		nu=(v.z/R.dir2.z-R.pos.z/R.dir2.z-v.x/R.dir1.y/R.dir2.z*R.dir1.z-R.pos.x/R.dir1.y/R.dir2.z*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.y/R.dir2.z*R.dir1.z);
		if (nu!=nu)
	{
		nu=(v.z/R.dir2.x-R.pos.z/R.dir2.x-v.x/R.dir1.z/R.dir2.x*R.dir1.z-R.pos.x/R.dir1.z/R.dir2.x*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.z/R.dir2.x*R.dir1.z);
		if (nu!=nu)
	{
		nu=(v.y/R.dir2.y-R.pos.z/R.dir2.y-v.x/R.dir1.z/R.dir2.y*R.dir1.z-R.pos.x/R.dir1.z/R.dir2.y*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.z/R.dir2.y*R.dir1.z);
		if (nu!=nu)
	{
		nu=(v.z/R.dir2.z-R.pos.z/R.dir2.z-v.x/R.dir1.z/R.dir2.z*R.dir1.z-R.pos.x/R.dir1.z/R.dir2.z*R.dir1.z)/(1.0f+R.dir2.x/R.dir1.z/R.dir2.z*R.dir1.z);
		}}}}}}}
	}
	labda=(v.x-R.pos.x-R.dir2.x*nu)/R.dir1.x;
	if (labda!=labda)
	{
		labda=(v.y-R.pos.y-R.dir2.y*nu)/R.dir1.y;
		if (labda!=labda)
	{
		labda=(v.z-R.pos.z-R.dir2.z*nu)/R.dir1.z;
		}
	}
	if (nu<0.0f||nu>1.0f||labda<0.0f||labda>1.0f)
		return false;
	return true;
}

CVector CSphere_Rect(CSphere S,CRect R)
{
	// Distance sphere to rectangle plane
	CVector a=R.pos-S.pos;
	float angle=CVectorAngle(a,R.normal);
	float dist=a.magnitude()*cos(angle);
	// Intersect plane R
	if (abs(dist)<S.radius)
	{
		R.normal.normalize();
		R.normal*=dist;
		CVertex v=S.pos-R.normal;
		// Inside or outside
		if (CVertex_Rect(v,R))
			return R.normal;
		else
			return CNullvect;
	}
	else
		return CNullvect;
}