/*
Copyright (c) 2012, David Liu
All rights reserved. Read the license for more information.
*/

/* Macros */
#if defined(COS_WIN32)
#define CMain int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int iCmdShow)
#elif defined(COS_LINUX)
#define CMain int main(int args,char**argc)
#endif

/* Keys */
#if defined(COS_WIN32)
// Letters
#define CKEY_A 0x41
#define CKEY_B 0x42
#define CKEY_C 0x43
#define CKEY_D 0x44
#define CKEY_E 0x45
#define CKEY_F 0x46
#define CKEY_G 0x47
#define CKEY_H 0x48
#define CKEY_I 0x49
#define CKEY_J 0x4A
#define CKEY_K 0x4B
#define CKEY_L 0x4C
#define CKEY_M 0x4D
#define CKEY_N 0x4E
#define CKEY_O 0x4F
#define CKEY_P 0x50
#define CKEY_Q 0x51
#define CKEY_R 0x52
#define CKEY_S 0x53
#define CKEY_T 0x54
#define CKEY_U 0x55
#define CKEY_V 0x56
#define CKEY_W 0x57
#define CKEY_X 0x58
#define CKEY_Y 0x59
#define CKEY_Z 0x5A
// Numbers
#define CKEY_0 0x30
#define CKEY_1 0x31
#define CKEY_2 0x32
#define CKEY_3 0x33
#define CKEY_4 0x34
#define CKEY_5 0x35
#define CKEY_6 0x36
#define CKEY_7 0x37
#define CKEY_8 0x38
#define CKEY_9 0x39
// Symbols
#define CKEY_SPACE 0x20
#define CKEY_BACKSPACE 0x08
#define CKEY_TAB 0x09
#define CKEY_ENTER 0x0D
#define CKEY_RIGHT 0x27
#define CKEY_LEFT 0x25
#define CKEY_DOWN 0x28
#define CKEY_UP 0x26
#define CKEY_SEMICOLON 0xBA
#define CKEY_DIVIDE 0xBF
#define CKEY_TILDE 0xC0
#define CKEY_RBRACKET 0xDD
#define CKEY_LBRACKET 0xDB
#define CKEY_SEPERATOR 0xDC
#define CKEY_QUOTE 0xDE
#define CKEY_MINUS 0xBD
#define CKEY_PLUS 0xBB
#define CKEY_PERIOD 0xBE
#define CKEY_COMMA 0xBC
// Modifier
#define CKEY_RSHIFT 0x10
#define CKEY_LSHIFT 0x10
#define CKEY_RCTRL 0x11
#define CKEY_LCTRL 0x11
#define CKEY_ESC 0x1B
// F-keys
#define CKEY_F1 0x70
#elif defined(COS_LINUX)
// Letters
unsigned char CKEY_A,CKEY_B,CKEY_C,CKEY_D,CKEY_E,CKEY_F,CKEY_G,CKEY_H,CKEY_I,CKEY_J,CKEY_K,CKEY_L,CKEY_M,CKEY_N
,CKEY_O,CKEY_P,CKEY_Q,CKEY_R,CKEY_S,CKEY_T,CKEY_U,CKEY_V,CKEY_W,CKEY_X,CKEY_Y,CKEY_Z,CKEY_0
,CKEY_1,CKEY_2,CKEY_3,CKEY_4,CKEY_5,CKEY_6,CKEY_7,CKEY_8,CKEY_9,CKEY_SPACE,CKEY_BACKSPACE,CKEY_TAB,CKEY_ENTER,CKEY_RIGHT,CKEY_LEFT,
CKEY_DOWN,CKEY_UP,CKEY_SEMICOLON,CKEY_DIVIDE,CKEY_TILDE,CKEY_RBRACKET,CKEY_LBRACKET
,CKEY_SEPERATOR,CKEY_QUOTE,CKEY_MINUS,CKEY_PLUS,CKEY_PERIOD,CKEY_COMMA,CKEY_RSHIFT,CKEY_LSHIFT,CKEY_RCTRL,CKEY_LCTRL,CKEY_ESC,CKEY_F1;
#endif
// Mouse
#define CKEY_LMOUSE 256
#define CKEY_RMOUSE 257

/* Functions */
#if defined(COS_WIN32)
bool CEngine::CSetupWindow(cwindowattribs*attribs)
{
	// Window type
	WNDCLASSEX wcex;
	wcex.cbSize=sizeof(WNDCLASSEX);
	wcex.style=CS_OWNDC;
	wcex.lpfnWndProc=CWndProc;
	wcex.cbClsExtra=0;
	wcex.cbWndExtra=0;
	wcex.hInstance=hInstance;
	strcpy(outerr,rootpath);
	if (!attribs->customicon)
		wcex.hIcon=NULL;
	else if ((wcex.hIcon=(HICON)LoadImage(NULL,strcat(outerr,"engine/rc_icon.ico"),IMAGE_ICON,0,0,LR_LOADFROMFILE))==NULL)
	{
		CMsgBox("Unable to find/load the icon file. Program will now close.","Error");
		return false;
	}
	strcpy(outerr,rootpath);
	if (!attribs->customcursor)
		wcex.hCursor=LoadCursor(NULL,IDC_ARROW);
	else if ((wcex.hCursor=(HCURSOR)LoadImage(NULL,strcat(outerr,"engine/rc_cursor.cur"),IMAGE_CURSOR,0,0,LR_LOADFROMFILE))==NULL)
	{
		CMsgBox("Unable to find/load the cursor file. Program will now close.","Error");
		return false;
	}
	wcex.hbrBackground=NULL;
	wcex.lpszMenuName=NULL;
	wcex.lpszClassName=attribs->wndname;
	wcex.hIconSm=NULL;
	// Registering window
	if(!RegisterClassEx(&wcex))
	{
		CMsgBox("Unable to register the window. Program will now close.","Error");
		return false;
	}
	// Creating window
	if (!(hWnd=CreateWindow(attribs->wndname,attribs->wndtitle,
		WS_POPUP|WS_VISIBLE|WS_CLIPSIBLINGS|WS_CLIPCHILDREN,xpos,ypos,width,height,
		NULL,NULL,hInstance,NULL)))
	{
		CMsgBox("Unable to create the window. Program will now close.","Error");
		return false;
	}
	SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)this);
	// Show it properly
	ShowWindow(hWnd,SW_SHOW);
	SetForegroundWindow(hWnd);
	SetFocus(hWnd);
	UpdateWindow(hWnd);
	return true;
}

LRESULT CALLBACK CEngine::CWndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	CEngine*receiver=(CEngine*)GetWindowLongPtr(hWnd,GWLP_USERDATA);
	if (receiver)
		return receiver->CProcWindow(message,wParam,lParam);
	return DefWindowProc(hWnd,message,wParam,lParam);
}

inline LRESULT CEngine::CProcWindow(UINT message,WPARAM wParam,LPARAM lParam)
{
	switch(message)
	{
	case WM_CLOSE:
		exit(1);
		break;
	case WM_ACTIVATE:
		{
			if (wParam!=WA_INACTIVE&&!focus)
			{
				CMinimize(false);
				focus=true;
			}
			else if (wParam==WA_INACTIVE&&focus)
			{
				CMinimize(true);
				focus=false;
			}
		}break;
	case WM_LBUTTONDOWN:
		key[CKEY_LMOUSE]=true;
		break;
	case WM_LBUTTONUP:
		key[CKEY_LMOUSE]=false;
		break;
	case WM_RBUTTONDOWN:
		key[CKEY_RMOUSE]=true;
		break;
	case WM_RBUTTONUP:
		key[CKEY_RMOUSE]=false;
		break;
	case WM_MOUSEMOVE:
		{
			mx=LOWORD(lParam);
			my=height-HIWORD(lParam);
		}break;
	case WM_KEYDOWN:
		key[wParam]=true;
		break;
	case WM_KEYUP:
		key[wParam]=false;
		break;
	default:
		return DefWindowProc(hWnd,message,wParam,lParam);
	}
	return 0;
}
#elif defined(COS_LINUX)
bool CEngine::CSetupWindow(cwindowattribs*attribs)
{
    int temp;
    int attrib[]={GLX_RGBA,GLX_DOUBLEBUFFER,GLX_RED_SIZE,4,GLX_GREEN_SIZE,4,GLX_BLUE_SIZE,4,GLX_DEPTH_SIZE,16,None};
    // Setting up GLX
    hDis=XOpenDisplay(NULL);
    if (hDis==NULL)
    {
        CMsgBox("Unable to open display. Program will now close","Error");
        return false;
    }
    if (!glXQueryExtension(hDis,&temp,&temp))
    {
        CMsgBox("X server does not have GLX extension support. Program will now close.","Error");
        return false;
    }
    hScr=DefaultScreen(hDis);
    hVi=glXChooseVisual(hDis,hScr,attrib);
    if (hVi==NULL)
    {
        CMsgBox("Unable to setup visual data. Program will now close","Error");
        return false;
    }
    hCx=glXCreateContext(hDis,hVi,0,true);
    if (hCx==NULL)
    {
        CMsgBox("Unable create OpenGL context. Program will now close","Error");
        return false;
    }
    cMap=XCreateColormap(hDis,RootWindow(hDis,hVi->screen),hVi->visual,AllocNone);
    // Setting properties
    maxwidth=DisplayWidth(hDis,hScr);
    maxheight=DisplayHeight(hDis,hScr);
    wAtt.colormap=cMap;
    wAtt.border_pixel=0;
    wAtt.event_mask=KeyPressMask|KeyReleaseMask|ButtonPressMask|ButtonReleaseMask|PointerMotionMask|FocusChangeMask;
    wAtt.override_redirect=true;
    if (width==-1)
    {
        width=maxwidth;
		height=maxheight;
    }
    XWMHints*wHints=XAllocWMHints();
    XClassHint*cHints=XAllocClassHint();
    XSizeHints*sHints=XAllocSizeHints();
    if (wHints==NULL||cHints==NULL||sHints==NULL)
    {
        CMsgBox("Unable to allocate WM hints. Program will now close","Error");
        return false;
    }
    XTextProperty wName,cName;
    XStringListToTextProperty(&attribs->wndtitle,1,&wName);
    XStringListToTextProperty(&attribs->wndname,1,&cName);
    wHints->initial_state=NormalState;
    wHints->input=false;
    wHints->flags=InputHint|StateHint;
    cHints->res_name=attribs->wndtitle;
    cHints->res_class=attribs->wndname;
    sHints->flags=PPosition|PSize;
     // Init keys
    CKEY_A=XKeysymToKeycode(hDis,XK_a);
    CKEY_B=XKeysymToKeycode(hDis,XK_b);
    CKEY_C=XKeysymToKeycode(hDis,XK_c);
    CKEY_D=XKeysymToKeycode(hDis,XK_d);
    CKEY_E=XKeysymToKeycode(hDis,XK_e);
    CKEY_F=XKeysymToKeycode(hDis,XK_f);
    CKEY_G=XKeysymToKeycode(hDis,XK_g);
    CKEY_H=XKeysymToKeycode(hDis,XK_h);
    CKEY_I=XKeysymToKeycode(hDis,XK_i);
    CKEY_J=XKeysymToKeycode(hDis,XK_j);
    CKEY_K=XKeysymToKeycode(hDis,XK_k);
    CKEY_L=XKeysymToKeycode(hDis,XK_l);
    CKEY_M=XKeysymToKeycode(hDis,XK_m);
    CKEY_N=XKeysymToKeycode(hDis,XK_n);
    CKEY_O=XKeysymToKeycode(hDis,XK_o);
    CKEY_P=XKeysymToKeycode(hDis,XK_p);
    CKEY_Q=XKeysymToKeycode(hDis,XK_q);
    CKEY_R=XKeysymToKeycode(hDis,XK_r);
    CKEY_S=XKeysymToKeycode(hDis,XK_s);
    CKEY_T=XKeysymToKeycode(hDis,XK_t);
    CKEY_U=XKeysymToKeycode(hDis,XK_u);
    CKEY_V=XKeysymToKeycode(hDis,XK_v);
    CKEY_W=XKeysymToKeycode(hDis,XK_w);
    CKEY_X=XKeysymToKeycode(hDis,XK_x);
    CKEY_Y=XKeysymToKeycode(hDis,XK_y);
    CKEY_Z=XKeysymToKeycode(hDis,XK_z);
    CKEY_0=XKeysymToKeycode(hDis,XK_0);
    CKEY_1=XKeysymToKeycode(hDis,XK_1);
    CKEY_2=XKeysymToKeycode(hDis,XK_2);
    CKEY_3=XKeysymToKeycode(hDis,XK_3);
    CKEY_4=XKeysymToKeycode(hDis,XK_4);
    CKEY_5=XKeysymToKeycode(hDis,XK_5);
    CKEY_6=XKeysymToKeycode(hDis,XK_6);
    CKEY_7=XKeysymToKeycode(hDis,XK_7);
    CKEY_8=XKeysymToKeycode(hDis,XK_8);
    CKEY_9=XKeysymToKeycode(hDis,XK_9);
    CKEY_SPACE=XKeysymToKeycode(hDis,XK_space);
    CKEY_BACKSPACE=XKeysymToKeycode(hDis,XK_space);
    CKEY_TAB=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_ENTER=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_RIGHT=XKeysymToKeycode(hDis,XK_Right);
    CKEY_LEFT=XKeysymToKeycode(hDis,XK_Left);
    CKEY_DOWN=XKeysymToKeycode(hDis,XK_Down);
    CKEY_UP=XKeysymToKeycode(hDis,XK_Up);
    CKEY_SEMICOLON=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_DIVIDE=XKeysymToKeycode(hDis,XK_slash);
    CKEY_TILDE=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_RBRACKET=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_LBRACKET=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_SEPERATOR=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_QUOTE=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_MINUS=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_PLUS=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_PERIOD=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_COMMA=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_RSHIFT=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_LSHIFT=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_RCTRL=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_LCTRL=XKeysymToKeycode(hDis,XK_semicolon);
    CKEY_ESC=XKeysymToKeycode(hDis,XK_Escape);
    CKEY_F1=XKeysymToKeycode(hDis,XK_semicolon);
    // Final window
    hWin=XCreateWindow(hDis,RootWindow(hDis,hVi->screen),xpos,ypos,width,height,0,hVi->depth,InputOutput,hVi->visual,
                        CWBorderPixel|CWColormap|CWEventMask|CWOverrideRedirect,&wAtt);
    // Create fake window
    XSetWindowAttributes tAtt;
    tAtt.event_mask=FocusChangeMask;
    fake=XCreateWindow(hDis,RootWindow(hDis,hVi->screen),-1,-1,1,1,0,0,InputOutput,0,CWEventMask,&tAtt);
    XSetWMProperties(hDis,hWin,&wName,&cName,NULL,0,sHints,wHints,cHints);
    // Icon
    unsigned int pw,ph;
    int px,py;
    strcpy(outerr,rootpath);
    if (attribs->customicon)
    {
        if (XpmReadFileToPixmap(hDis,hWin,strcat(outerr,"engine/rc_icon.xpm"),&hIcon,NULL,NULL)!=XpmSuccess)
        {
            CMsgBox("Unable find/load the icon pixmap. Program will now close","Error");
            hIcon=NULL;
            return false;
        }
    }
    else
        hIcon=NULL;
    // Fake properties
    wHints->icon_pixmap=hIcon;
    wHints->initial_state=NormalState;
    wHints->input=true;
    wHints->flags=InputHint|StateHint|IconPixmapHint;
    cHints->res_name=attribs->wndtitle;
    cHints->res_class=attribs->wndname;
    sHints->flags=PPosition|PSize;
    XSetWMProperties(hDis,fake,&wName,&cName,NULL,0,sHints,wHints,cHints);
    XFree(wHints);
    XFree(cHints);
    XFree(sHints);
    // Setup the cursors
    Pixmap pempty;
    XColor black;
    char nodata[]={0,0,0,0,0,0,0,0};
    black.red=black.green=black.blue=0;
    pempty=XCreateBitmapFromData(hDis,hWin,nodata,8,8);
    hide=XCreatePixmapCursor(hDis,pempty,pempty,&black,&black,0,0);
    strcpy(outerr,rootpath);
    if (attribs->customcursor)
    {
        if (XpmReadFileToPixmap(hDis,hWin,strcat(outerr,"engine/rc_cursor.xpm"),&hCursor,NULL,NULL)!=XpmSuccess)
        {
            CMsgBox("Unable find/load the icon pixmap. Program will now close","Error");
            hCursor=NULL;
            return false;
        }
    }
    else
        hCursor=NULL;
    // Show window
    glXMakeCurrent(hDis,hWin,hCx);
    XMapWindow(hDis,fake);
    XMapRaised(hDis,hWin);
    XSetInputFocus(hDis,hWin,RevertToNone,CurrentTime);
    return true;
}

inline void CEngine::CProcessMessage()
{
    while(XPending(hDis)>0)
    {
        XNextEvent(hDis,&hEvent);
        switch (hEvent.type)
        {
        case KeyPress:
            key[hEvent.xkey.keycode]=true;
            break;
        case KeyRelease:
            key[hEvent.xkey.keycode]=false;
            break;
        case ButtonPress:
            {
                switch(hEvent.xbutton.button)
                {
                case 1:
                    key[CKEY_LMOUSE]=true;
                    break;
                case 3:
                    key[CKEY_RMOUSE]=true;
                    break;
                }
            }break;
        case ButtonRelease:
            {
                switch(hEvent.xbutton.button)
                {
                case 1:
                    key[CKEY_LMOUSE]=false;
                    break;
                case 3:
                    key[CKEY_RMOUSE]=false;
                    break;
                }
            }break;
        case FocusIn:
            {
                if (hEvent.xfocus.window==fake&&!focus)
                {
                    XRaiseWindow(hDis,hWin);
                    XSetInputFocus(hDis,hWin,RevertToNone,CurrentTime);
                    focus=true;
                }
            }break;
        case FocusOut:
            {
                if (hEvent.xfocus.window==hWin&&focus)
                {
                    XLowerWindow(hDis,hWin);
                    focus=false;
                }
            }break;
        case MotionNotify:
            {
                mx=hEvent.xmotion.x;
                my=height-hEvent.xmotion.y;
            }break;
        }
    }
}
#endif
