/* Headers */
#include "CEngine.h"
#include "Definitions.h"
#include "Parser.h"
#include "Benchmark.h"
#include "Main.h"

// Main
bool Init()
{
	// OpenGL
	glClearColor(0,0,0,1);
	core.CResize2D(core.width,core.height,10.0f,10.0f);
	core.CVerticalSync(false);
	glDepthFunc(GL_LEQUAL);
	glCullFace(GL_BACK);
	// Loading frame
	glLoadIdentity();
	glTranslatef(0.0f,0.0f,-1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	core.CPrintText("Loading...",1.0f,1.0f,CGuicol_w,CFONT_TEXT);
	core.CSwapBuffers();
	// Userfile
	unsigned char curline=0;
	char lines[1][2];
	strcpy(core.outerr,core.rootpath);
	ifstream userfile(strcat(core.outerr,"data/userfile.cfg"));
	if (userfile.is_open())
	{
		while (userfile.good())
		{
			userfile >> lines[curline];
			curline++;
		}
		userfile.close();
	}
	else
	{
		core.CMsgBox("Unable to open/find userfile.cfg. Program will now close.","Error");
		return false;
	}
	// Find benchmarks
	app.benchcnt=core.CSearchDir("benchmarks/","*.obs",&app.benchnames); // only look for .obs files
	if (app.benchcnt==-1)
	{
		core.CMsgBox("Unable to search the benchmark directory. Program will now close.","Error");
		return false;
	}
	// Setup scripter
	SetupScripter();
	// Setup
	app.quit=false;
	app.barcol[0]=0.7f;app.barcol[1]=1.0f;app.barcol[2]=1.0f;
	app.extscrl=app.benchscrl=0.0f;
	if (atoi(lines[0])==0)
	{
		core.CSoundVol(CVOL_MUTE);
		app.opt1=false;
		strcpy(app.str1,"Off");
	}
	else
	{
		core.CSoundVol(CVOL_MAX);
		app.opt1=true;
		strcpy(app.str1,"On");
	}
	memset(app.extfind,0,sizeof(app.extfind));
	app.extpos=0;
	app.extclick=false;
	app.pmx=-1;
	app.rot=0.0f;
	app.benchmode=false;
	app.prevtime=app.lefttime=0;
	// OpenGL strings
	unsigned int count;
	count=strlen((char*)glGetString(GL_VERSION))+1;
	gldata.glversion=new char[count];
	strcpy(gldata.glversion,(char*)glGetString(GL_VERSION));
	count=strlen((char*)glGetString(GL_SHADING_LANGUAGE_VERSION))+1;
	if (count>1)
	{
		gldata.glslversion=new char[count];
		strcpy(gldata.glslversion,(char*)glGetString(GL_SHADING_LANGUAGE_VERSION));
	}
	else
	{
		count++;
		gldata.glslversion=new char[count];
		strcpy(gldata.glslversion,"-");
	}
	count=strlen((char*)glGetString(GL_VENDOR))+1;
	gldata.implementor=new char[count];
	#if defined(COS_LINUX)
	int maj,min;
	// Use temp buffer storage
	glXQueryVersion(core.hDis,&maj,&min);
	count=sprintf(gldata.implementor,"%d.%d",maj,min)+1;
	gldata.glxversion=new char[count];
	strcpy(gldata.glxversion,gldata.implementor);
	// Continue
	#else
	gldata.glxversion=new char[2];
	strcpy(gldata.glxversion,"-");
	#endif
	strcpy(gldata.implementor,(char*)glGetString(GL_VENDOR));
	count=strlen((char*)glGetString(GL_RENDERER))+1;
	gldata.renderer=new char[count];
	strcpy(gldata.renderer,(char*)glGetString(GL_RENDERER));
	// Extensions
	gldata.extlen=strlen((char*)glGetString(GL_EXTENSIONS))+1;
	if (gldata.extlen>1)
	{
		char*extstr=new char[gldata.extlen];
		strcpy(extstr,(char*)glGetString(GL_EXTENSIONS));
		unsigned int i=0,sav=0,loop,abslp,cnt=0;
		gldata.extcnt=0;
		while (i<gldata.extlen)
		{
			if (extstr[i]==' ')
				gldata.extcnt++;
			i++;
		}
		gldata.extensions=new char*[gldata.extcnt];
		i=0;
		// Get sizes and allocate array(s)
		while (i<gldata.extlen)
		{
			if (extstr[i]==' ')
			{
				loop=i-sav+1;
				gldata.extensions[cnt]=new char[loop];
				sav=++i;
				cnt++;
			}
			i++;
		}
		cnt=sav=i=0;
		// Get all seperate extensions into seperate arrays
		while (i<gldata.extlen)
		{
			if (extstr[i]==' ')
			{
				abslp=0;
				CLoopc(loop,sav,i,1)
				{
					gldata.extensions[cnt][abslp]=extstr[loop];
					abslp++;
				}
				gldata.extensions[cnt][abslp]='\0';
				sav=++i;
				cnt++;
			}
			i++;
		}
		delete[] extstr;
	}
	else
		gldata.extcnt=0;
	// Preparing
	core.CSetCursor(core.width/2.0,core.height/2.0);
	return true;
}

void Main()
{
	if (app.benchmode)
	{
		if (core.focus)
		{
			while ((app.lefttime=core.CGetTime()-app.prevtime)<bench.frame)
			{
				// Render
				BenchRender((float)app.lefttime/bench.frame);
			}
			app.prevtime=core.CGetTime()-(app.lefttime-bench.frame);
			// Process
			BenchProcess();
		}
		else // Pause
		{
            core.CSleep(APPFRAME);
            app.prevtime=core.CGetTime()-app.lefttime;
		}
	}
	else
	{
		if (app.prevtime!=0)
		{
			app.lefttime=core.CGetTime()-app.prevtime;
			app.lefttime=APPFRAME-app.lefttime;
		}
		if (app.lefttime>0) // Time left
		{
			core.CSleep(app.lefttime);
			app.prevtime=core.CGetTime();
		}
		else // Too slow
			app.prevtime=core.CGetTime()+app.lefttime;
        if (core.focus)
        {
            if (app.page==P_MENU)
                Menu();
            else if (app.page==P_DATA)
                Data();
            else if (app.page==P_OPTIONS)
                Options();
            else if (app.page==P_ABOUT)
                About();
            else if (app.page==P_MARKS)
                Benchmarks();
            else
			Extensions();
        }
	}
}

void Quit()
{
	// Memory
	int cnt;
	delete[] gldata.glversion;
	delete[] gldata.glslversion;
	delete[] gldata.implementor;
	delete[] gldata.renderer;
	delete[] gldata.glxversion;
	if (app.benchcnt>0)
	{
		CLoops(cnt,app.benchcnt)
		{
			delete[] app.benchnames[cnt];
		}
		delete[] app.benchnames;
	}
	if (gldata.extcnt>0)
	{
		CLoops(cnt,gldata.extcnt)
		{
			delete[] gldata.extensions[cnt];
		}
		delete[] gldata.extensions;
	}
}

CMain
{
	// Init
	cwindowattribs attributes;
	attributes.type=CWIN_POPUP;
	attributes.wndname="Wnd";
	attributes.wndtitle="OpenGLChecker";
	attributes.customcursor=false;
	attributes.customicon=true;
	if (!core.CSetupEngine(&attributes,APPWIDTH,APPHEIGHT)||!Init())
		app.quit=true;

	// Main
	while (!app.quit)
	{
		core.CProcessMessage();
		Main();
	}

	// Quit
	Quit();
	core.CCleanupEngine();
	return 0;
}
